"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
const MAX_LIMIT_VALUE = 0x0fffffff | 0;
const LIMIT_REGEX = /Database returned more than the allowed response size limit/i;
const RETRY_LIMIT = Symbol('RETRY_LIMIT');

const retrieveRows = async (pool, fieldList, query, currentSkip, currentLimit) => {
  try {
    const {
      inlineLedgerExecuteStatement,
      convertResultRow
    } = pool;
    const inputRows = await inlineLedgerExecuteStatement(pool, `${query} OFFSET ${currentSkip} LIMIT ${currentLimit}`, inlineLedgerExecuteStatement.SHARED_TRANSACTION_ID);
    const rows = [];

    for (let idx = 0; idx < inputRows.length; idx++) {
      rows[idx] = convertResultRow(inputRows[idx], fieldList);
    }

    return rows;
  } catch (error) {
    if (error == null || !(LIMIT_REGEX.test(error.message) || LIMIT_REGEX.test(error.stack) || LIMIT_REGEX.test(error.code))) {
      throw error;
    }

    throw RETRY_LIMIT;
  }
};

const makeRowsLimitError = () => new Error('Database returned more than the allowed response size limit');

const find = async (pool, readModelName, tableName, searchExpression, fieldList, sort, inputSkip, inputLimit) => {
  const {
    searchToWhereExpression,
    makeNestedPath,
    inlineLedgerExecuteStatement,
    escapeId,
    escapeStr,
    tablePrefix,
    schemaName
  } = pool;
  const orderExpression = sort && Object.keys(sort).length > 0 ? 'ORDER BY ' + Object.keys(sort).map(fieldName => {
    const [baseName, ...nestedPath] = fieldName.split('.');
    const provisionedName = nestedPath.length === 0 ? escapeId(baseName) : `${escapeId(baseName)}->'${makeNestedPath(nestedPath)}'`;
    return sort[fieldName] > 0 ? `${provisionedName} ASC` : `${provisionedName} DESC`;
  }).join(', ') : '';
  const searchExpr = searchToWhereExpression(searchExpression, escapeId, escapeStr, makeNestedPath);
  const inlineSearchExpr = searchExpr.trim() !== '' ? `WHERE ${searchExpr} ` : '';
  const query = `SELECT * FROM ${escapeId(schemaName)}.${escapeId(`${tablePrefix}${tableName}`)}
  ${inlineSearchExpr}
  ${orderExpression}
  `;
  const skip = isFinite(+inputSkip) ? +inputSkip : 0;
  let limit = isFinite(+inputLimit) ? +inputLimit : MAX_LIMIT_VALUE;

  try {
    return await retrieveRows(pool, fieldList, query, skip, limit);
  } catch (error) {
    if (error !== RETRY_LIMIT) {
      throw error;
    }
  }

  try {
    limit = (await inlineLedgerExecuteStatement(pool, `WITH "CTE" AS (${query} OFFSET ${skip})
      SELECT Count("CTE".*) AS "Count" FROM "CTE"
      `, inlineLedgerExecuteStatement.SHARED_TRANSACTION_ID))[0].Count;
  } catch (error) {
    throw makeRowsLimitError();
  }

  for (let [factor, size] = [2, Math.ceil(limit / 2)]; !Number.isNaN(factor); void ([factor, size] = [factor < limit ? Math.ceil(limit / (factor * 2)) * (factor * 2) - limit <= Math.ceil(limit / (factor * 2)) ? factor * 2 : limit : Number.NaN, Math.ceil(limit / (factor * 2))])) {
    try {
      return Array.from([]).concat(...(await Promise.all(Array.from({
        length: factor
      }).map((_, index) => retrieveRows(pool, fieldList, query, skip + index * size, index < factor - 1 ? size : limit - (factor - 1) * size)))));
    } catch (error) {
      if (error !== RETRY_LIMIT) {
        throw RETRY_LIMIT;
      }
    }
  }

  throw makeRowsLimitError();
};

var _default = find;
exports.default = _default;
//# sourceMappingURL=find.js.map